--[[
	Allows you to keep track of people you have had whisper conversations with.
]]

-- Variables
local TellTrack_ID = 0
local TellTrack_ButtonCount = 1
--local TellTrack_ArrayMaxSize = 20

local TellTrack_WhisperMessages = { }
local TellTrack_CurrentConversation = { name = "", showAll = true, }
local TellTrack_OfflinePlayers = { }

-- hooked funtions
local SavedSendChatMessage = nil
local SavedFCF_SelectDockFrame = nil
local SavedChatFrame_OpenChat = nil
local SavedFCF_Close = nil
local SavedFCF_OpenNewWindow = nil

SELECTED_CHAT_FRAME = DEFAULT_CHAT_FRAME
ERR_CHAT_PLAYER_NOT_FOUND_S_REGEX = gsub(ERR_CHAT_PLAYER_NOT_FOUND_S, "%%s", "(.*)")
ERR_FRIEND_OFFLINE_S_REGEX = gsub(ERR_FRIEND_OFFLINE_S, "%%s", "(.*)")
ERR_FRIEND_ONLINE_SS_REGEX = gsub(ERR_FRIEND_ONLINE_SS, "%%s", "(.*)")

-- executed on load, calls general set-up functions
function TellTrack:OnLoadCustom( frame )

	SLASH_TELLTRACK3 = "/ttrack"
	SlashCmdList["TELLTRACKRETELL"] = TellTrack_WhisperToPreviousTarget
	SLASH_TELLTRACKRETELL1 = "/re"
	SLASH_TELLTRACKRETELL2 = "/retell"
        
	frame:RegisterEvent("PLAYER_ENTERING_WORLD")
	frame:RegisterEvent("CHAT_MSG_WHISPER")
	frame:RegisterEvent("CHAT_MSG_WHISPER_INFORM")
	frame:RegisterEvent("CHAT_MSG_SYSTEM")
	
	if SendChatMessage ~= SavedSendChatMessage then							-- Hook the chat event handler 'before'
		SavedSendChatMessage = SendChatMessage
		SendChatMessage = TellTrack_SendChatMessage
	end
	if FCF_SelectDockFrame ~= TellTrack_FCF_SelectDockFrame then			-- Hook dock frame selection 'after'
		SavedFCF_SelectDockFrame = FCF_SelectDockFrame
		FCF_SelectDockFrame = TellTrack_FCF_SelectDockFrame
	end
	if ChatFrame_OpenChat ~= TellTrack_ChatFrame_OpenChat then				-- Hook open chat to open as whisper
		SavedChatFrame_OpenChat = ChatFrame_OpenChat
		ChatFrame_OpenChat = TellTrack_ChatFrame_OpenChat
	end
	if FCF_Close ~= TellTrack_FCF_Close then								-- Hook open chat to open as whisper
		SavedFCF_Close = FCF_Close
		FCF_Close = TellTrack_FCF_Close
	end
	if FCF_OpenNewWindow ~= TellTrack_FCF_OpenNewWindow then				-- Hook open chat to open as whisper
		SavedFCF_OpenNewWindow = FCF_OpenNewWindow
		FCF_OpenNewWindow = TellTrack_FCF_OpenNewWindow
	end
	
	self._cmdlist = {
		enabled = { name="enabled", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_Enabled) end, help=TELLTRACK_CHAT_COMMAND_ENABLE_INFO},
		invert = { name="invert", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_Inverted) end, help=TELLTRACK_CHAT_COMMAND_INVERT_INFO},
		dontsavelist = { name="dontsavelist", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_DontSaveList) end, help=TELLTRACK_CHAT_COMMAND_DONTSAVELIST_INFO},
		autocreate = { name="autocreate", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_AutoCreate) end, help=TELLTRACK_CHAT_COMMAND_AUTOCREATE_INFO},
		hidewhispers = { name="hidewhispers", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_HideOtherWhispers) end, help= TELLTRACK_CHAT_COMMAND_HIDEOTHERWHISPERS_INFO},
		timestamps = { name="timestamps", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_TimeStamps) end, help=TELLTRACK_CHAT_COMMAND_TIMESTAMPS_INFO},
		whisperfirst = { name="whisperfirst", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_WhisperFirst) end, help=TELLTRACK_CHAT_COMMAND_WHISPERFIRST_INFO},
		clearall = { name="clearall", func=TellTrack_Menu_DeleteAll, help=TELLTRACK_CHAT_COMMAND_CLEARALL_INFO},
		border = { name="border", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_HideBorder) end, help=TELLTRACK_CHAT_COMMAND_HIDEBORDER_INFO},
		linenums = { name="linenums", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_LineNums) end, help=TELLTRACK_CHAT_COMMAND_LINENUMS_INFO},
		showall = { name="showall", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_ShowAll) end, help=TELLTRACK_CHAT_COMMAND_SHOWALL_INFO},
	}
end

function TellTrack:OnVariablesLoaded( frame )
	local db,o = self._db,self._options

	self.L = self._acelib.L
	self.LR = self._acelib.LR
	
	o.Realm,o.Race,o.Player,o.Faction = GetCVar("realmName"),UnitRace("player"),self:UnitName("Player"),"Alliance" 
	for k,v in pairs({"Orc","Tauren","Troll","Undead"}) do if o.Race == v then o.Faction = "Horde"; break; end end
	o.ListProfile = o.Realm .. " - " .. o.Faction
	o.PlayerProfile = o.Player .. " - " .. o.Realm

--	getglobal(frame:GetName() .. "HeaderIcon"):SetTexture( self._meta.icon )
--	getglobal(frame:GetName() .. "HeaderTitle"):SetText( self._meta.title )
--	getglobal(frame:GetName() .. "HeaderVersion"):SetText( self._options.Color .. "v" .. self._meta.version )  -- .. " "..self._meta.locale.."|r" )

	-- load frame list
	self._chat = db.Chat[o.PlayerProfile] or { }
	db.Chat[o.PlayerProfile] = self._chat
	
	-- load player list
	TellTrack_Toggle_Enabled(o.Enabled)
	if o.DontSaveList == 1 then db.Log = { } end
	self._list = db.Log[o.ListProfile] or { }
	db.Log[o.ListProfile] = self._list
	self:UpdateTellTrackButtonsText()
	
	TellTrack_Toggle_UpdateBorderVisibility()
	UIDropDownMenu_Initialize(TellTrackDropDown, TellTrackTextButton_InitializeMenuArray, "MENU")	--Initialize Menu
end

function TellTrack:OnPlayerLogin( frame )
	local _inst = self._acelib
	if _inst and _inst.menu then
		_inst.menu.args.help = nil
	end
end

function TellTrack:OnCommandCustom( msg, cmd, args )
	local cmdlist = self._cmdlist
	if not cmd then
		cmdlist["enabled"].func(args); return true
	elseif cmdlist[cmd] then
		cmdlist[cmd].func(args); return true
	elseif cmd == "help" then
		self:Print(TELLTRACK_CHAT_USE_INFO)
		for k,v in pairs(cmdlist) do
			self:Print("/"..self._meta.short.." "..k.." - ".. v.help)
		end
	end
end

function TellTrack:Print_Help()
	self:Print(TELLTRACK_CHAT_QUESTION_MARK_INFO)
	self:Print(TELLTRACK_CHAT_QUESTION_MARK_INFO1)
	self:Print(TELLTRACK_CHAT_QUESTION_MARK_INFO2)
	self:Print(TELLTRACK_CHAT_QUESTION_MARK_INFO3)
end


function TellTrack:ChatCommandToggler(msg, func)
	msg = msg and strlower(strtrim(msg)) or ""
	
	-- Toggle appropriately
	if ( (string.find(msg, 'on')) or ((string.find(msg, '1')) and (not string.find(msg, '-1')) ) ) then
		func(1)
	elseif ( (string.find(msg, 'off')) or (string.find(msg, '0')) ) then
		func(0)
	else
		func(-1)
	end
end

-- Toggles the enabled/disabled state of an option and returns the new state
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack:Generic_Toggle(toggle, variableName, enableMessage, disableMessage)
	local oldvalue = self._options[variableName]
	local newvalue = toggle
	if ( toggle ~= 1 ) and ( toggle ~= 0 ) then
		if oldvalue == 1 then
			newvalue = 0
		elseif oldvalue == 0 then
			newvalue = 1
		else
			newvalue = 0
		end
	end
	self._options[variableName] = newvalue
	if newvalue ~= oldvalue then
		if variableName~="Enabled" then
			if newvalue == 1 then
				self:Print(getglobal(enableMessage))
			else
				self:Print(getglobal(disableMessage))
			end
		end
	end
	return newvalue
end

-- Toggles the enabled/disabled state of the TellTrack
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_Enabled(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "Enabled", "TELLTRACK_CHAT_ENABLED", "TELLTRACK_CHAT_DISABLED")
	if  TellTrack._options.Enabled == 0 then
		TellTrackFrame:Hide()
		if TellTrackTooltip then
			TellTrackTooltip:Hide()
		end
	else
		TellTrackFrame:Show()
	end
end

-- Toggles the inverted/normalized state of the TellTrack list
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_Inverted(toggle)
	local tt = TellTrack
	local o = tt._options
	local oldvalue = o.InvertedList
	local newvalue = tt:Generic_Toggle(toggle, "InvertedList", "TELLTRACK_CHAT_INVERTED", "TELLTRACK_CHAT_NORMALIZED")
	--only change (reset offset) telltrack if InvertedList was changed
	if oldvalue ~= newvalue then
		tt:ChangeArrayOffset((newvalue == 1) and o.ListSize or 1)
		tt:UpdateTellTrackButtonsText()
	end
end

-- Toggles the saved state of the TellTrack list accross sessions
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_DontSaveList(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "DontSaveList", "TELLTRACK_CHAT_DONTSAVELIST", "TELLTRACK_CHAT_SAVELIST")
end

-- Toggles the auto creation of TellTrack whisper frame
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_AutoCreate(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "AutoCreate", "TELLTRACK_CHAT_AUTOCREATE", "TELLTRACK_CHAT_MANUALWHISPERFRAME")
end

-- Toggles Hiding Whispers outside the whisper frame
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_HideOtherWhispers(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "HideOtherWhispers", "TELLTRACK_CHAT_HIDEOTHERWHISPERS", "TELLTRACK_CHAT_SHOWOTHERWHISPERS")
	if ( newvalue == 1 ) then
		TellTrack:HideWhispersInOtherFrames()
	end
end

-- Toggles Whisper chat stamping
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_TimeStamps(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "TimeStamps", "TELLTRACK_CHAT_TIMESTAMPS", "TELLTRACK_CHAT_NOTIMESTAMPS")
	TellTrack:RedrawWhisperFrame()
end

-- Toggles the whisper first option
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_WhisperFirst(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "WhisperFirst", "TELLTRACK_CHAT_WHISPERFIRST", "TELLTRACK_CHAT_LOGFIRST")
end

-- Toggles the border hiding option
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_HideBorder(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "HideBorder", "TELLTRACK_CHAT_HIDEBORDER", "TELLTRACK_CHAT_SHOWBORDER")
	TellTrack_Toggle_UpdateBorderVisibility()
end

-- Toggles the line number hiding option
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_LineNums(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "LineNums", "TELLTRACK_CHAT_SHOWLINENUMS", "TELLTRACK_CHAT_HIDELINENUMS")
	TellTrack:UpdateTellTrackButtonsText()
end

function TellTrack_Toggle_ShowAll()
	local newvalue = TellTrack:Generic_Toggle(toggle, "ShowAll", "TELLTRACK_CHAT_SHOWALL", "TELLTRACK_CHAT_HIDEOTHER")
	TellTrack_ShowAllWhispers()
end
function TellTrack_ShowAllWhispers()
	TellTrack:ShowConversationWith(TellTrack_CurrentConversation.name, true)
end

function TellTrack_Toggle_UpdateBorderVisibility()
	if (TellTrack._options.HideBorder == 1) then
		TellTrackFrameBackground:Hide()
		TellTrackFrameArrowUpButton:Hide()
		TellTrackFrameArrowDownButton:Hide()
		TellTrackFrame:SetHitRectInsets(10, 10, 10+TellTrackFrameArrowUpButton:GetHeight(), 10+TellTrackFrameArrowDownButton:GetHeight())
	else
		TellTrackFrameBackground:Show()
		TellTrackFrameArrowUpButton:Show()
		TellTrackFrameArrowDownButton:Show()
		TellTrackFrame:SetHitRectInsets(0, 0, 0, 0)
	end
end

function TellTrack_WhisperToPreviousTarget(msg)
	local o = TellTrack._options
	if not o.LastTell then return end
	SendChatMessage(msg, "WHISPER", this.language, o.LastTell)
end

function TellTrack_Menu_Whisper()
	TellTrack:InitiateWhisperByID(TellTrack_ID)
end

function TellTrack_Menu_Show_Conversation()
	TellTrack:ShowConversationWith(TellTrack:GetName(TellTrack_ID), false)
end

function TellTrack_Menu_Who()
	local name = TellTrack:GetName(TellTrack_ID)
	SendWho("n-\""..name.."\"")
end

function TellTrack_Menu_GroupeInvite()
	local name = TellTrack:GetName(TellTrack_ID)
	InviteUnit(name)
end

function TellTrack_Menu_AddToFriend()
	local name = TellTrack:GetName(TellTrack_ID)
	AddFriend(name)
end

function TellTrack_Menu_Delete()
	TellTrack:EraseByID(TellTrack_ID)
end

-- Function by AnduinLothar used in right click menu
function TellTrack_Menu_DeleteAll()
	local tt = TellTrack
	local db,o = tt._db,tt._options
	tt._list = {}; db.Log[o.ListProfile] = tt._list
	TellTrack:UpdateTellTrackButtonsText()
end


-- Handles events
function TellTrack:OnEventCustom(frame, event, ...)
--	if ( event == "PLAYER_ENTERING_WORLD" ) then
--		self:ModifyButtonCount(TellTrackFrame:GetHeight())
	if event == "CHAT_MSG_WHISPER" or event == "CHAT_MSG_WHISPER_INFORM" then
		--For some reason self:HookAllAddMessages doesn't work on VARIABLES_LOADED...
		self:HookAllAddMessages()
		
		--Special Hidden Whisper Ignores
        if ( IsAddOnLoaded("Guilded") ) and (string.len(arg1) > 2) and (string.sub(arg1,1,2) == "<G") and ( (string.len(arg1) ~= 4) or (string.sub(arg1,1,4) ~= "<GM>") ) then
			return				--Hide Guilded messages
		elseif ( IsAddOnLoaded("GuildAds") ) and (string.len(arg1) > 3) and (string.sub(arg1,1,3) == "<GA") then
			return				--Hide GuildAds messages
		elseif (string.len(arg1) > 3) and (string.sub(arg1,1,3) == "<T>") then
			return				--Hide Telepathy messages
        elseif (string.len(arg1) > 1) and (string.sub(arg1,1,1) == "/") then
            return				--Hide AceComm messages
        elseif (string.len(arg1) > 7) and (string.sub(arg1,1,7) == "SKP1019") then
            return				--Hide SKP messages
		end 
		
		local type = string.sub(event,10)
		if event == "CHAT_MSG_WHISPER" then
			self:HandleMessageSentOrRecieved(arg2, false)
		end
		
		local message = gsub(arg1, "%%", "%%%%")
        if arg2 then
            TellTrack_OfflinePlayers[arg2] = nil
        end
		table.insert(TellTrack_WhisperMessages, {type=type, tag=arg6, name=arg2, msg=message, time=self:GetTime(nil, self._update.timediff)})
	elseif ( event == "CHAT_MSG_SYSTEM" ) then
		local _, _, name = strfind(arg1, ERR_CHAT_PLAYER_NOT_FOUND_S_REGEX)
		if (name) then
			TellTrack_OfflinePlayers[name] = true
			self:UpdateTellTrackButtonsText()
			return
		end
		_, _, name = strfind(arg1, ERR_FRIEND_OFFLINE_S_REGEX)
		if (name) then
			TellTrack_OfflinePlayers[name] = true
			self:UpdateTellTrackButtonsText()
			return
		end
		_, _, name = strfind(arg1, ERR_FRIEND_ONLINE_SS_REGEX)
		if (name) then
			TellTrack_OfflinePlayers[name] = nil
			self:UpdateTellTrackButtonsText()
			return
		end
	end
end



-- adds a whisper to a frame

function TellTrack:GetTimeStamp(ts)
	local separator = "" -- ">"
	if ChatTimeStamps_TimeStamp and not ChatTimeStamps_SEPARATOR then
		separator = ""
	end
	if not ts then
		if ChatTimeStamps_LOCALTIME then
			ts = date()
		elseif Clock_GetTimeText then
			ts = Clock_GetTimeText()
		else
			ts = self:GetTime(nil, self._update.timediff)
		end
	end
	if type(ts) == "number" then
		ts = date("%H:%M:%S", ts)
	end	
	if type(ts) ~= "string" then
		ts = ""
	end
	return ts..separator.." "
end

function TellTrack:RedrawWhisperFrame( )	
	local frame = self._chat.frame and getglobal(self._chat.frame)
	if not frame then return end
	local scrollOffset,found,current = frame:GetCurrentScroll(),false,TellTrack_CurrentConversation
	frame:Clear()
	
	--Display only the valid whispers.
	for i,v in pairs(TellTrack_WhisperMessages) do
		if current.showAll or (strlower(v.name) == strlower(current.name)) then
			local info = ChatTypeInfo[v.type]
			local part1 = getglobal("CHAT_"..v.type.."_GET")..v.msg
			local part2 = (v.tag and _G["CHAT_FLAG_"..v.tag] or "").."|Hplayer:"..v.name.."|h".."["..v.name.."]".."|h"
			if TasteTheNaimbow_Loaded then
				part2 = "|cff"..TasteTheNaimbowExternalColor(name)..part2.."|r"
			end
			frame:AddMessage(format(part1, part2), info.r, info.g, info.b, info.id, nil, v.time)
			found = true
		end
	end
	
	frame:SetScrollOffset(scrollOffset)
	return found
end

function TellTrack:ShowConversationWith(name, showAll)
	local db,o,chat = self._db,self._options,self._chat
	local frame = chat.frame and getglobal(chat.frame)
	if not chat.frame or not getglobal(chat.frame) then
		if o.AutoCreate==0 or not self:CreateWhisperChatFrame() then
			return
		end
	end
	
	TellTrack_CurrentConversation = { name=name, showAll=showAll, }
	if not self:RedrawWhisperFrame() then
		TellTrack_CurrentConversation.showAll = true
		local timeStamp = o.TimeStamps
		o.TimeStamps = 0
		frame:AddMessage(format(TELLTRACK_NO_CURRENT_CONVERSATION, name))
		o.TimeStamps = timeStamp
	end
	chat.name = name
	
	-- Show the frame and tab
	FCF_SetWindowName(frame, chat.name)
    if frame.isDocked then
		FCF_SelectDockFrame(frame)
    end
end


function TellTrack_ChatFrame_AddMessage(frame, text, r, g, b, id, addToStart, ts)
	local tt = TellTrack
	local AddMessage = SavedChatFrame_AddMessages[frame:GetName()]
	
	if frame:GetName() ~= tt._chat.frame then
		AddMessage(frame, text, r, g, b, id, addToStart)
		return					-- incorrect whisper frame
	end
	
	local stamps = ChatTimeStamps_ENABLED
	ChatTimeStamps_ENABLED = 0
	if strlen(text) < 5 then
		AddMessage(frame, text, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
		return
	end
	local timedText = text
	if (tt._options.TimeStamps == 1) then
		timedText = tt:GetTimeStamp(ts)..text
	end
	if (not TellTrack_CurrentConversation) or (TellTrack_CurrentConversation.showAll) then
		AddMessage(frame, timedText, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
		return
	end
	
	local name = TellTrack_CurrentConversation.name
    local namecolor = "ffffff"
    if TasteTheNaimbow_Loaded and TasteTheNaimbowExternalColor then
        namecolor = self:SafeCall(TasteTheNaimbowExternalColor, name)
    end
    --DEFAULT_CHAT_FRAME:AddMessage("Real "..string.gsub(string.gsub(text,":%d+\124","\124"),"\124","\\124"))
    --DEFAULT_CHAT_FRAME:AddMessage("Fake "..string.gsub(format(CHAT_WHISPER_GET, "|Hplayer:"..name.."|h".."["..name.."]".."|h"),"\124","\\124"))
    --DEFAULT_CHAT_FRAME:AddMessage(tostring((tt:startsWith(string.gsub(text,":%d+\124","\124"), format(CHAT_WHISPER_GET, "|Hplayer:"..name.."|h".."["..name.."]".."|h")))))
    local matchText = strlower(string.gsub(text,":%d+\124","\124"))  --remove numbers
    local Hplayer = "|Hplayer:"..name.."|h".."["..name.."]".."|h"
    local HplayerL = "[".."|Hplayer:"..name.."|h"..name.."|h".."]"
    local function nameMatch(get, name)
		local s, s2 = matchText, strlower(format(get, name));
		if type(s) ~= "string" or type(s2) ~= "string" then return false end		
		return s == s2 or (s:len() > s2:len() and string.sub(s, 1, s2:len()) == s2)
	end
	if	nameMatch(CHAT_WHISPER_GET, Hplayer) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_DND..Hplayer) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_AFK..Hplayer) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_GM..Hplayer) or
		nameMatch(CHAT_WHISPER_INFORM_GET, Hplayer) or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..Hplayer.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..Hplayer.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..Hplayer.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..Hplayer.."|r") or
		nameMatch(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..Hplayer.."|r") or
		nameMatch(CHAT_WHISPER_GET, HplayerL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_DND..HplayerL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_AFK..HplayerL) or
		nameMatch(CHAT_WHISPER_GET, CHAT_FLAG_GM..HplayerL) or
		nameMatch(CHAT_WHISPER_INFORM_GET, HplayerL) or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..HplayerL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND..HplayerL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK..HplayerL.."|r") or
		nameMatch(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM..HplayerL.."|r") or
		nameMatch(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor..HplayerL.."|r") then
		AddMessage(frame, timedText, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
		return
	end
end


function TellTrack:HookAllAddMessages()
	if SavedChatFrame_AddMessages then
		return
	end
	SavedChatFrame_AddMessages = {}
	for i=1, NUM_CHAT_WINDOWS do
		local frameName = "ChatFrame"..i
		local frame = getglobal(frameName)
		if frame and (frame.AddMessage ~= TellTrack_ChatFrame_AddMessage) then
			SavedChatFrame_AddMessages[frameName] = frame.AddMessage
			frame.AddMessage = TellTrack_ChatFrame_AddMessage
		end
	end
end
		
function TellTrack:FindChatFrame( ... )
	for i=1, NUM_CHAT_WINDOWS do
		local sChatFrame = "ChatFrame"..i
		local chatFrame = getglobal(sChatFrame)
		local sName, _, _, _, _, _, bShown = GetChatWindowInfo(i)
		if select("#",...) == 0 then
			if (not bShown and not chatFrame.isDocked) or i == NUM_CHAT_WINDOWS then
				return chatFrame
			end	
		elseif bShown or chatFrame.isDocked then
			for i1 = 1, select("#",...) do
				local name = select(i1,...)
				if name and name == sName then
					return sChatFrame,sName
				end
			end
		end
	end
end 

function TellTrack:CreateWhisperChatFrame()
	local chat,default,name = self._chat,TELLTRACK_WHISPER_FRAME_TITLE
	if chat.frame and getglobal(chat.frame) then return chat.frame end
	
	chat.frame,name = self:FindChatFrame(chat.name, chat.name~=default and default);	-- get if already exists
	if not chat.frame then
		if FCF_GetNumActiveChatFrames() >= NUM_CHAT_WINDOWS then return end				-- dont get if have to take someone elses
		name = chat.name or default
		local selected = SELECTED_DOCK_FRAME
		FCF_OpenNewWindow(name)
		FCF_SelectDockFrame(selected)
		chat.frame = self:FindChatFrame(name)
	end
	if chat.frame then																	-- make sure clear out
		chat.name = name
		local chatFrame = getglobal(chat.frame)
		ChatFrame_RemoveAllMessageGroups(chatFrame)
		ChatFrame_AddMessageGroup(chatFrame, "WHISPER")
		return true
	end
end

function TellTrack:HideWhispersInOtherFrames()
	local chat = self._chat
	if not chat.frame or not getglobal(chat.frame) then	return end
	for i=1, NUM_CHAT_WINDOWS do
		local sChatFrame = "ChatFrame"..i
		local sName, _, _, _, _, _, bShown = GetChatWindowInfo(i)
		if (bShown or chatFrame.isDocked) and sChatFrame ~= chat.frame then
			ChatFrame_RemoveMessageGroup(getglobal(sChatFrame), "WHISPER")
		end
	end
end


-- Does things with the hooked function
function TellTrack_SendChatMessage(text, type, language, target)
	TellTrack:SendChatMessage(text, type, language, target)
end

function TellTrack:SendChatMessage(text, type, language, target)
	local o = self._options
	if type == "WHISPER" then
		o.LastTell = target
		if o.Enabled == 1 and strsub(text, 1, 1) ~= "<" then	-- prevent data message transfers from being used
			self:HandleMessageSentOrRecieved(target, true)
		end
	end
	if SavedSendChatMessage then
		SavedSendChatMessage(text, type, language, target)
	end
end

function TellTrack_ChatFrame_OpenChat(text, chatFrame)
	local chat = TellTrack._chat
	local cc = TellTrack_CurrentConversation
	if (text == "") and (SELECTED_CHAT_FRAME:GetName() == chat.frame) and (cc)
			and (cc.name) and (cc.name ~= "") then
		TellTrack:InitiateWhisperToTarget(cc.name)
	else
		if SavedChatFrame_OpenChat then
			SavedChatFrame_OpenChat(text, chatFrame)
		end
	end
end

function TellTrack_FCF_OpenNewWindow( name )
	local isHide = (TellTrack._options.HideOtherWhispers == 1)
	local chatFrame = (isHide) and TellTrack:FindChatFrame() or nil
	if SavedFCF_OpenNewWindow then
		SavedFCF_OpenNewWindow(name)
		if isHide and chatFrame then
			ChatFrame_RemoveMessageGroup(chatFrame, "WHISPER")
		end
	end
end

function TellTrack_FCF_Close( chatFrame, fallBack )
	local frame = fallBack or chatFrame or FCF_GetCurrentChatFrame()
	if frame and frame:GetName() == TellTrack._chat.frame then
		TellTrack._chat.frame = nil
	end
	if SavedFCF_Close then
		SavedFCF_Close(chatFrame, fallBack)
	end
end

function TellTrack_FCF_SelectDockFrame(frame)
	local tt = TellTrack
	local o,list = tt._options,tt._list
	
	if SavedFCF_SelectDockFrame then
		SavedFCF_SelectDockFrame(frame)
	end
	if frame:GetName() == tt._chat.frame then
		if TellTrack_CurrentConversation.showAll then
			for i = 1, o.ListSize do
				if list[i] then	list[i].unread = 0 end
			end
		elseif TellTrack_CurrentConversation.name ~= "" then
			for i = 1, o.ListSize do
				if list[i] and strlower(list[i].name) == strlower(TellTrack_CurrentConversation.name) then
					list[i].unread = 0
				end
			end
		end
		tt:UpdateTellTrackButtonsText()
	end
	SELECTED_CHAT_FRAME = frame;		--FIXME: is this necessary
end

function TellTrack:ArrowButton_OnClick( frame, button, isUp )
	local o = self._options
	if not frame:GetID() then return end
	if button == "RightButton" then
		self:ChangeArrayOffset( isUp and 1 or o.ListSize )
	elseif button == "LeftButton" then
		self:ChangeArrayOffset( o.ListOffset + (TellTrack_ButtonCount * (isUp and -1 or 1)) )
	else
		return
	end
	self:UpdateTellTrackButtonsText()
end

function TellTrack:ArrowButton_OnSizeChanged( frame )
	local height = frame:GetHeight()
	getglobal(frame:GetName().."RightBG"):SetHeight(height)
	getglobal(frame:GetName().."LeftBG"):SetHeight(height)
	getglobal(frame:GetName().."Arrow"):SetHeight(height)
end


function TellTrack:InitiateWhisperByID(id)
	local name = self:GetName(id)
	self:InitiateWhisperToTarget(name)
end

-- 1/20/05 'Retell' added by AnduinLothar
function TellTrack:InitiateWhisperToPreviousTarget()
	local o = self._options
	if not o.LastTell then return end
	self:InitiateWhisperToTarget(o.LastTell)
end

function TellTrack:InitiateWhisperToTarget(name)
	if not name then return end
	
	local chatFrame = DEFAULT_CHAT_FRAME
	chatFrame.editBox:SetAttribute("chatType", "WHISPER")
	chatFrame.editBox:SetAttribute("tellTarget", name)
	ChatEdit_UpdateHeader(chatFrame.editBox)
	if not chatFrame.editBox:IsVisible() then
		if SavedChatFrame_OpenChat then
			SavedChatFrame_OpenChat("", chatFrame)
		end
	end
end

function TellTrackTextButton_InitializeMenuArray( frame, level )
	TellTrack._rcmenu = TellTrack._rcmenu or {
		{ text = "   "..BINDING_HEADER_TELLTRACKHEADER, isTitle = 1, },
		{ text = TELLTRACK_WHISPER, func = TellTrack_Menu_Whisper, },
		{ text = TELLTRACK_SHOW_CONVERSATION, func = TellTrack_Menu_Show_Conversation, },
		{ text = TELLTRACK_WHO, func = TellTrack_Menu_Who, },
		{ text = TELLTRACK_GRPINV, func = TellTrack_Menu_GroupeInvite, },
		{ text = TELLTRACK_ADDFRIEND, func = TellTrack_Menu_AddToFriend, },
		{ text = TELLTRACK_DELETE, func = TellTrack_Menu_Delete, },
		{ text = "|cffcccccc--------------|r", disabled = 1, notClickable = 1, },
		{ text = TELLTRACK_DELETE_ALL, func = TellTrack_Menu_DeleteAll, },
		{ text = TELLTRACK_INVERT, func = TellTrack_Toggle_Inverted, },
		{ text = TELLTRACK_SHOW_ALL_WHISPERS, func = TellTrack_ShowAllWhispers, },
		{ text = TELLTRACK_CREATE_CHATFRAME, func = function(name) TellTrack:CreateWhisperChatFrame(name) end, },
		{ text = "|cffcccccc--------------|r", disabled = 1, notClickable = 1, },
		{ text = TELLTRACK_CANCEL, func = function() end, },
	}
	for i,v in pairs(TellTrack._rcmenu) do
		UIDropDownMenu_AddButton(v)
	end
end

function TellTrack:TextButton_OnClick(frame, button)
	local id = frame:GetID()
	if not id then return end
	if self._list[self:GetArrayId(id)] == nil then
		return
	end
	TellTrack_ID = id
	if button == "RightButton" then
		ToggleDropDownMenu(1, nil, TellTrackDropDown, "cursor", 0, 0)
	elseif button == "LeftButton" then
		if (IsShiftKeyDown() and IsAltKeyDown()) then
			TellTrack_Menu_Delete()
		elseif (IsShiftKeyDown()) then
			TellTrack_Menu_Who()
		elseif (IsAltKeyDown()) then
			TellTrack_Menu_GroupeInvite()
		elseif (IsControlKeyDown()) then
			TellTrack_Menu_AddToFriend()
		else
			local name = self:GetName(id)
			if (self._options.WhisperFirst == 1) then
				if (SELECTED_DOCK_FRAME.editBox:IsVisible()) and (SELECTED_DOCK_FRAME.editBox:GetAttribute("chatType") == "WHISPER") and (strlower(SELECTED_DOCK_FRAME.editBox:GetAttribute("tellTarget")) == strlower(name)) then
					self:ShowConversationWith(name, false)
				else
					self:InitiateWhisperToTarget(name)
				end
			else
				if (SELECTED_DOCK_FRAME == getglobal(self._chat.frame)) and (strlower(TellTrack_CurrentConversation.name) == strlower(name)) then
					self:InitiateWhisperToTarget(name)
				else
					self:ShowConversationWith(name, false)
				end
			end
		end
	end
end

function TellTrack:GetArrayId( id )
	local o = self._options
	if not id then return end
	if o.InvertedList ~= 1 then
		id = (id + o.ListOffset) - 1
	else
		id = (o.ListSize - (id + o.ListOffset)) + 2
	end
	return id
end

function TellTrack:EraseByID( id )
	if id then
		id = self:GetArrayId(id)
		if self._list and self._list[id] then
			if TellTrackTooltip:IsVisible() and 0 == id then
				TellTrackTooltip:Hide()
			end
			tremove(self._list, id)
		end
		self:UpdateTellTrackButtonsText()
	end
end

function TellTrack:GetName(id)
	local list = self._list
	if not id or not list then return end
	id = self:GetArrayId(id)
	if list[id] and list[id].name then
		return list[id].name
	end
end

function TellTrack:SetTooltip(id)
	local name = self:GetName(id)
	if name then TellTrackTooltip:SetText(name) end
end

function TellTrack:UpdateTellTrackButtonsText()
	if not self._list then return end 
	local o = self._options
	local maxitems = TellTrack_ButtonCount
	local lastToFmtStr = "|cff33ff33%s|r"
	local lastFromFmtStr = "|cffff3333%s|r"
	local noNameFmtStr = "|cff666666%s|r"

	for i=1, maxitems do
		local id = self:GetArrayId(i)
		local sline = "TellTrack" .. i
		local line = getglobal(sline.."Text")
		local item = self._list[id]
		local fmt, val = noNameFmtStr, "Empty"
		
		if not line then
		else
			if item and item.dispName then
				val = item.dispName
				if not TellTrack_OfflinePlayers[val] then
					fmt = (item.sentTo) and lastToFmtStr or lastFromFmtStr
				end
				if item.unread and item.unread > 0 then
					val = "("..item.unread..") "..val
				end
				if o.LineNums == 1 then
					val = id .. ". "..val
				end
			end
			line:SetText(format(fmt, val))
			line:Show()
		end
	end
end

--[[
	1/15/05 - AnduinLothar: The last person whispered is automaticly moved to the bottom of
	the list and everytime you whisper it autoscrolls to the bottom of the list. Also if the
	list is full the top name is deleted and the new name is appended to the bottom.    
	
	1/17/05 - Added recieved/sent boolean for color on button reload.
]]--
function TellTrack:CapUTF8( s )
	if not s then return end
	local s1,len,c = "",s:len()
	for i = 1, len do
		c = s:byte(i);
		if i==1 and c<=127 then s1=s1..strupper(strchar(c));
		elseif c<=127 then	s1=s1..strlower(strchar(c));
		else				s1=s1..strchar(c); end
	end
	return s1
end

function TellTrack:HandleMessageSentOrRecieved( target, isSent )
	if not target then return end
	local o = self._options
	local cmpName,dispName = strlower(target),self:CapUTF8(target)
	local max, current = o.ListSize, TellTrack_CurrentConversation
	
	-- find
	local item, id = nil, 0
	for k, v in pairs(self._list) do
		if v and v.cmpName == cmpName then item,v = v,nil; end
		if v then id = id + 1; if id ~= k then self._list[id] = v; end end			-- compress as we go
	end
	while (id >= max) do tremove(self._list, 1); id = id - 1; end
	if not item then item = { name = target, cmpName = cmpName, dispName = dispName, } end
	id = id + 1; self._list[id] = item
	
	-- update
	item.sentTo = isSent
	if not item.unread then	item.unread = 0	end
	if not isSent and self._chat.frame then
		if (not getglobal(self._chat.frame):IsVisible()) or (current.name ~= target and not current.showAll) then
			item.unread = item.unread + 1
		end
	end
	
	-- change offset
	if self._options.InvertedList == 1 then
		if id < max - TellTrack_ButtonCount and (max - id + 1) ~= o.ListOffset then
			self:ChangeArrayOffset(max-id+1)
		end
	else
		if id > TellTrack_ButtonCount and (id-TellTrack_ButtonCount+1) ~= o.ListOffset then
			self:ChangeArrayOffset(id-TellTrack_ButtonCount+1)
		end
	end

	self:UpdateTellTrackButtonsText()
end


function TellTrack:ChangeArrayOffset(offset)
	local o = self._options
	if offset <= 0 then
		offset = 1
	end
	local capSize = o.ListSize - TellTrack_ButtonCount + 1
	if offset > capSize then
		offset = capSize
	end
	o.ListOffset = offset
end


function TellTrack:OnMouseWheel(frame, v)
	local o = self._options
	local v = v < 0 and 1 or (v > 0 and -1 or 0)
	self:ChangeArrayOffset(o.ListOffset -  v)
	self:UpdateTellTrackButtonsText()
end


function TellTrack:OnSizeChanged( frame )
	local o = self._options
	local width = frame:GetWidth() - 24
	TellTrackFrameArrowUpButton:SetWidth(width)
	for i=1, o.ListSize do
		getglobal("TellTrack"..i):SetWidth(width)
	end
	TellTrackFrameArrowDownButton:SetWidth(width)

	local height = frame:GetHeight() - 44				--insets and arrow buttons
	local count = floor(height / 20)
	if count ~= TellTrack_ButtonCount then
		if self._list then
			local max, id = TellTrack_ArrayMaxSize, #self._list
			local oh = TellTrack_ButtonCount - id; oh = (oh < 0) and 0 or oh;
			if self._options.InvertedList ~= 1 then
				o.ListOffset = (o.ListOffset + (TellTrack_ButtonCount - count)) - oh;
			else
				o.ListOffset = id + oh
			end
			self:ChangeArrayOffset(o.ListOffset)
		end
		TellTrack_ButtonCount = count
		for i=1, o.ListSize do
			local sline = "TellTrack"..i
			local line = getglobal(sline)
			if i > TellTrack_ButtonCount then line:Hide() else line:Show() end
		end
	end
	local offset = floor((height - TellTrack_ButtonCount * 20 + 20) / 2)
	TellTrackFrameArrowUpButton:SetHeight(offset)
	TellTrackFrameArrowDownButton:SetHeight(offset)

	self:UpdateTellTrackButtonsText()
end

